"""Tests for the SmartApp file."""

import pytest

from pysmartapp.errors import SignatureVerificationError
from pysmartapp.smartapp import SmartApp

from .utilities import get_fixture


class TestSmartApp:
    """Tests for the SmartApp class."""

    @staticmethod
    def test_initialize():
        """Tests the property initialization."""
        # Arrange
        name = "Test Name"
        description = "Test Description"
        perms = ["Perm1"]
        app_id = "MyApp"
        # Act
        app = SmartApp(name, description, perms, app_id)
        # Assert
        assert app.name == name
        assert app.description == description
        assert app.permissions == perms
        assert app.app_id == app_id

    @staticmethod
    def test_ping():
        """Tests the ping lifecycle event."""
        # Arrange
        request = get_fixture("ping_request")
        expected_response = get_fixture("ping_response")
        app = SmartApp("Test Name", "Test Description", [])
        # Act
        response = app.handle_request(request)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_ping():
        """Tests the ping event handler."""
        # Arrange
        request = get_fixture("ping_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert app == smartapp
        smartapp.on_ping += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_config_init():
        """Tests the configuration initialization lifecycle event."""
        # Arrange
        request = get_fixture("config_init_request")
        expected_response = get_fixture("config_init_response")
        app = SmartApp("SmartApp", "SmartApp Description",
                       ['l:devices'], "myapp")
        # Act
        response = app.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_config_page():
        """Tests the configuration initialization page event."""
        # Arrange
        request = get_fixture("config_page_request")
        expected_response = get_fixture("config_page_response")
        app = SmartApp("SmartApp", "SmartApp Description",
                       ['l:devices'], "myapp")
        # Act
        response = app.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_config():
        """Tests the config event handler."""
        # Arrange
        request = get_fixture("config_init_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert app == smartapp
        smartapp.on_config += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_install():
        """Tests the install lifecycle event."""
        # Arrange
        request = get_fixture("install_request")
        expected_response = get_fixture("install_response")
        smartapp = SmartApp("SmartApp", "SmartApp Description",
                            ['l:devices'], "myapp")
        # Act
        response = smartapp.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_install():
        """Tests the install event handler."""
        # Arrange
        request = get_fixture("install_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert req.auth_token == '580aff1f-f0f1-44e0-94d4-e68bf9c2e768'
            assert req.refresh_token == 'ad58374e-9d6a-4457-8488-a05aa8337ab3'
            assert app == smartapp
        smartapp.on_install += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_update():
        """Tests the update lifecycle event."""
        # Arrange
        request = get_fixture("update_request")
        expected_response = get_fixture("update_response")
        app = SmartApp("SmartApp", "SmartApp Description",
                       ['l:devices'], "myapp")
        # Act
        response = app.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_update():
        """Tests the update event handler."""
        # Arrange
        request = get_fixture("update_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert req.auth_token == '4ebd8d9f-53b0-483f-a989-4bde30ca83c0'
            assert req.refresh_token == '6e3bbf5f-b68d-4250-bbc9-f7151016a77f'
            assert app == smartapp
        smartapp.on_update += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_event():
        """Tests the event lifecycle event."""
        # Arrange
        request = get_fixture("event_request")
        expected_response = get_fixture("event_response")
        app = SmartApp("SmartApp", "SmartApp Description",
                       ['l:devices'], "myapp")
        # Act
        response = app.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_event():
        """Tests the event event handler."""
        # Arrange
        request = get_fixture("event_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert app == smartapp
        smartapp.on_event += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_oauth_callback():
        """Tests the oauth_callback lifecycle event."""
        # Arrange
        request = get_fixture("oauth_callback_request")
        expected_response = get_fixture("oauth_callback_response")
        app = SmartApp("SmartApp", "SmartApp Description",
                       ['l:devices'], "myapp")
        # Act
        response = app.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_oauth_callback():
        """Tests the oauth_callback event handler."""
        # Arrange
        request = get_fixture("oauth_callback_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert app == smartapp
        smartapp.on_oauth_callback += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_uninstall():
        """Tests the uninstall lifecycle event."""
        # Arrange
        request = get_fixture("uninstall_request")
        expected_response = get_fixture("uninstall_response")
        app = SmartApp("SmartApp", "SmartApp Description",
                       ['l:devices'], "myapp")
        # Act
        response = app.handle_request(request, None, False)
        # Assert
        assert response == expected_response

    @staticmethod
    def test_on_uninstall():
        """Tests the uninstall event handler."""
        # Arrange
        request = get_fixture("uninstall_request")
        smartapp = SmartApp("Test Name", "Test Description", [])
        fired = False

        def handler(req, resp, app):
            nonlocal fired
            fired = True
            assert app == smartapp
        smartapp.on_uninstall += handler
        # Act
        smartapp.handle_request(request, None, False)
        # Assert
        assert fired

    @staticmethod
    def test_handle_request_sig_verification():
        """Tests handle_request with sig verification."""
        # Arrange
        public_key = get_fixture('public_key', 'pem')
        data = get_fixture('config_init_sig_pass_request')
        smartapp = SmartApp("Test Name", "Test Description", [],
                            public_key=public_key)
        # Act
        resp = smartapp.handle_request(data['body'], data['headers'], True)
        # Assert
        assert resp

    @staticmethod
    def test_handle_request_sig_verification_missing_headers():
        """Tests handle_request with sig verification."""
        # Arrange
        public_key = get_fixture('public_key', 'pem')
        data = get_fixture('config_init_sig_pass_request')
        smartapp = SmartApp("Test Name", "Test Description", [],
                            public_key=public_key)
        # Act/Assert
        with pytest.raises(SignatureVerificationError):
            smartapp.handle_request(data['body'], [], True)

    @staticmethod
    def test_handle_request_sig_verification_fails():
        """Tests handle_request with sig verification."""
        # Arrange
        public_key = get_fixture('public_key', 'pem')
        data = get_fixture('config_init_sig_fail_request')
        smartapp = SmartApp("Test Name", "Test Description", [],
                            public_key=public_key)
        # Act/Assert
        with pytest.raises(SignatureVerificationError):
            smartapp.handle_request(data['body'], data['headers'], True)



"""Testing utilities."""

import json


def get_fixture(file: str, ext: str = 'json'):
    """Load a fixtures file."""
    file_name = "tests/fixtures/{file}.{ext}".format(file=file, ext=ext)
    with open(file_name) as open_file:
        if ext == 'json':
            return json.load(open_file)
        return open_file.read()



""""Tests for the uninstall module."""

from pysmartapp.consts import LIFECYCLE_UNINSTALL
from pysmartapp.uninstall import UninstallRequest

from .utilities import get_fixture


class TestUninstallRequest:
    """Tests for the UninstallRequest class."""

    @staticmethod
    def test_init():
        """Tests the init method."""
        # Arrange
        data = get_fixture('uninstall_request')
        # Act
        req = UninstallRequest(data)
        # Assert
        assert req.uninstall_data_raw == data['uninstallData']
        assert req.lifecycle == LIFECYCLE_UNINSTALL
        assert req.execution_id == data['executionId']
        assert req.locale == data['locale']
        assert req.version == data['version']
        assert req.installed_app_id == 'd692699d-e7a6-400d-a0b7-d5be96e7a564'
        assert req.location_id == 'e675a3d9-2499-406c-86dc-8a492a886494'
        assert req.installed_app_config == \
            data['uninstallData']['installedApp']['config']
        assert req.settings == data['settings']



"""Define a SmartApp."""

import logging
from typing import Sequence

from .eventhook import EventHook
from .utilities import create_request

_LOGGER = logging.getLogger(__name__)


class SmartApp:
    """Define the SmartApp class."""

    def __init__(self, name: str, description: str,
                 permissions: Sequence[str], app_id: str = 'app',
                 path: str = '/', public_key=None):
        """Initialize the SmartApp class."""
        self._name = name
        self._description = description
        self._permissions = permissions
        self._app_id = app_id
        self._path = path
        self._public_key = public_key
        self.on_ping = EventHook()
        self.on_config = EventHook()
        self.on_install = EventHook()
        self.on_update = EventHook()
        self.on_event = EventHook()
        self.on_oauth_callback = EventHook()
        self.on_uninstall = EventHook()

    def handle_request(self, data: dict, headers: list = None,
                       validate_signature: bool = True) -> dict:
        """Process a lifecycle event."""
        req = create_request(data)
        resp = req.process(self, headers, validate_signature)

        if req.installed_app_id:
            _LOGGER.debug("%s: %s received for installed app %s.",
                          req.execution_id, req.lifecycle,
                          req.installed_app_id)
        else:
            _LOGGER.debug("%s: %s received.",
                          req.execution_id, req.lifecycle)

        return resp.to_data()

    @property
    def name(self) -> str:
        """Get the name of the SmartApp."""
        return self._name

    @property
    def description(self) -> str:
        """Get the description of the SmartApp."""
        return self._description

    @property
    def permissions(self) -> Sequence[str]:
        """Get the install permissions of the SmartApp."""
        return self._permissions

    @property
    def app_id(self):
        """Get the app id of the SmartApp."""
        return self._app_id

    @property
    def path(self) -> str:
        """Get the path the SmartApp is installed at."""
        return self._path

    @property
    def public_key(self):
        """Get the public key of the SmartApp used to verify events."""
        return self._public_key



"""Define the event hook module."""


class EventHook:
    """Define the EventHook subscriber bag."""

    def __init__(self):
        """Initialize a new instance of the EventHook class."""
        self.__handlers = []

    def __iadd__(self, handler):
        """Add an event handler."""
        self.__handlers.append(handler)
        return self

    def __isub__(self, handler):
        """Remove an event handler."""
        self.__handlers.remove(handler)
        return self

    def fire(self, *args, **keywargs):
        """Invoke the event handlers."""
        for handler in self.__handlers:
            handler(*args, **keywargs)

    @property
    def handlers(self):
        """Get the bound handlers."""
        return self.__handlers



"""Tests for the EventHook file."""

from pysmartapp.eventhook import EventHook


class TestEventHook:
    """Tests for the EventHook class."""

    @staticmethod
    def test_add():
        """Tests adding an event handler."""
        hook = EventHook()
        hook += TestEventHook.__handler
        assert TestEventHook.__handler in hook.handlers

    @staticmethod
    def test_remove():
        """Tests removing an event handler."""
        hook = EventHook()
        hook += TestEventHook.__handler
        hook -= TestEventHook.__handler
        assert TestEventHook.__handler not in hook.handlers

    @staticmethod
    def test_fire():
        """Tests firing an event handler."""
        hook = EventHook()
        fired = False

        def handler():
            nonlocal fired
            fired = True

        hook += handler
        hook.fire()
        assert fired

    @staticmethod
    def __handler():
        pass
